/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * This file is part of maemo-video-thumbnailer package
 * A plugin for Tumbler to create video thumbnails
 *
 * Copyright (C) 2009 Nokia Corporation.  All Rights reserved.
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include "gst-thumb-pipe.h"

#include <string.h>
#include <utime.h>
#include <setjmp.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>
#include <glib-object.h>
#include <gio/gio.h>
#include <sys/file.h>

#include <gst/gst.h>

#include <jpeglib.h>

#define THUMBER_PIPE_ERROR_DOMAIN "ThumberPipeError"
#define SEEK_TIMEOUT 5
#define PIPE_TIMEOUT 10
#define VALID_VARIANCE_THRESHOLD  256.0
#define FRAME_REJECT_LIMIT 100

struct tej_error_mgr {
        struct jpeg_error_mgr jpeg;
        jmp_buf setjmp_buffer;
};

static void           newpad_callback                  (GstElement       *decodebin,
                                                        GstPad           *pad,
                                                        gboolean          last,
                                                        ThumberPipe      *pipe);

static gboolean       stream_continue_callback         (GstElement    *bin,
                                                        GstPad        *pad,
                                                        GstCaps       *caps,
                                                        ThumberPipe   *pipe);

static void           pad_notify_caps_callback         (GstPad *pad,
                                                        GParamSpec *pspec,
                                                        gpointer *user_data);

static void           buffer_callback                  (GstElement    *bin,
                                                        GstBuffer     *buffer,
                                                        GstPad        *pad,
                                                        ThumberPipe   *pipe);

static gboolean       wait_for_state_change            (ThumberPipe *pipe,
                                                        GstState     state,
                                                        GError     **error);

static gboolean       wait_for_image_buffer            (ThumberPipe *pipe,
                                                        const gchar *uri,
                                                        GError     **error);

static gchar *        filename_for_thumb               (const gchar *uri,
                                                        const gchar *flavor);

static gboolean       create_thumbnail                 (unsigned char *buffer,
                                                        const gchar *uri,
                                                        const guint width,
                                                        const guint height,
                                                        const guint rotation,
                                                        const ThumberFlavor *flavor);

static gboolean       check_for_valid_thumb            (unsigned char *buffer,
                                                        const gint     size);


static void           set_caps_filter                  (ThumberPipe *pipe,
                                                        GstCaps *caps);

G_DEFINE_TYPE (ThumberPipe, thumber_pipe, G_TYPE_OBJECT)

#define THUMBER_PIPE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), TYPE_THUMBER_PIPE, ThumberPipePrivate))

enum {
        ROTATE0,
        ROTATE90,
        ROTATE180,
        ROTATE270
};

typedef struct {
        GstElement     *pipeline;
        
        GstElement     *source;
        GstElement     *decodebin;

        GstElement     *sinkbin;
        GstElement     *video_scaler;
        GstElement     *video_color;
        GstElement     *video_filter;
        GstElement     *video_sink;

        gchar          *uri;
        const ThumberFlavor  *flavor;
        guint           width;
        guint           height;
        guint           rotation;

        guint           rejected_frame_count;
} ThumberPipePrivate;

enum {
	PROP_0,
};

enum {
	NO_ERROR,
	INITIALIZATION_ERROR,
	RUNNING_ERROR,
	THUMBNAIL_ERROR
};

static GQuark
error_quark (void)
{
	return g_quark_from_static_string (THUMBER_PIPE_ERROR_DOMAIN);
}

ThumberPipe *
thumber_pipe_new ()
{
	return g_object_new (TYPE_THUMBER_PIPE, NULL);
}

static void
thumber_pipe_set_property (GObject      *object,
                           guint         prop_id,
                           const GValue *value,
                           GParamSpec   *pspec)
{
	switch (prop_id) {
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}


static void
thumber_pipe_get_property (GObject    *object,
                           guint       prop_id,
                           GValue     *value,
                           GParamSpec *pspec)
{
	switch (prop_id) {
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}

static void
thumber_pipe_finalize (GObject *object)
{
	ThumberPipePrivate *priv;
	priv = THUMBER_PIPE_GET_PRIVATE (object);

	gst_element_set_state (priv->pipeline, GST_STATE_NULL);
	/* State changes to NULL are synchronous */
	gst_object_unref (priv->pipeline);

	G_OBJECT_CLASS (thumber_pipe_parent_class)->finalize (object);
}

static void
thumber_pipe_class_init (ThumberPipeClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	object_class->finalize     = thumber_pipe_finalize;
	object_class->set_property = thumber_pipe_set_property;
	object_class->get_property = thumber_pipe_get_property;

	g_type_class_add_private (object_class, sizeof (ThumberPipePrivate));
}

static void
thumber_pipe_init (ThumberPipe *object)
{

}

gboolean
thumber_pipe_run (ThumberPipe         *pipe,
                  const gchar         *uri,
                  const ThumberFlavor *flavor,
                  GError             **error)
{
	ThumberPipePrivate *priv;
	GstBuffer          *buffer = NULL;
	gchar              *filename;
	gboolean            success = FALSE;
	GError             *lerror  = NULL;
	gint64              duration = 0;
	gint64              seek;
	GstFormat           format = GST_FORMAT_TIME;

	priv = THUMBER_PIPE_GET_PRIVATE (pipe);

	g_return_val_if_fail (pipe != NULL, FALSE);
	g_return_val_if_fail (uri != NULL, FALSE);

	/*g_debug ("Running thumbnailing pipeline for uri: %s", uri);*/

	filename = g_filename_from_uri (uri, NULL, NULL);

	g_object_set (priv->source,
	              "location",
	              filename,
	              NULL);
	g_free (filename);

	priv->uri = g_strdup (uri);
	priv->flavor = flavor;
	priv->width = 0;
	priv->height = 0;
        priv->rotation = ROTATE0;
	priv->rejected_frame_count = 0;

	gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
	if (!wait_for_state_change (pipe, GST_STATE_PAUSED, &lerror)) {
		g_propagate_error (error, lerror);
		success = FALSE;
		goto cleanup;
	}

	if (!gst_element_query_duration (priv->pipeline, &format, &duration))
		goto skip_seek;

	if (strcmp(thumber_flavor_get_frame_selection (flavor), "First") == 0) {
		goto skip_seek;
	}

	if (duration > 120 * GST_SECOND) {
		seek = 45 * GST_SECOND;
	} else {
		seek = duration/3;
	}

	if (gst_element_seek (priv->pipeline, 1.0, GST_FORMAT_TIME,
	                      GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT,
	                      GST_SEEK_TYPE_SET, seek,
	                      GST_SEEK_TYPE_NONE, GST_CLOCK_TIME_NONE)) {
							  /* Wait for the seek to finish */
							  gst_element_get_state (priv->pipeline, NULL, NULL,
							                         SEEK_TIMEOUT * GST_SECOND);
						  }

 skip_seek:
	
	gst_element_set_state (priv->pipeline, GST_STATE_PLAYING);
	if (!wait_for_image_buffer (pipe, uri, &lerror)) {
		g_propagate_error (error, lerror);
		success = FALSE;
		goto cleanup;
	}
	
	success = TRUE;
	
	gst_element_set_state (priv->pipeline, GST_STATE_READY);
	
	gst_element_get_state (priv->pipeline, NULL, NULL,
			       SEEK_TIMEOUT * GST_SECOND);
	
 cleanup:
	gst_element_set_state (priv->pipeline, GST_STATE_NULL);
	
	g_free (priv->uri);
	
	return success;
}

gboolean
thumber_pipe_initialize (ThumberPipe *pipe, GError **error)
{
	ThumberPipePrivate *priv;
	GstPad             *videopad;
	GstCaps            *caps;

	priv = THUMBER_PIPE_GET_PRIVATE (pipe);

	priv->pipeline     = gst_pipeline_new ("source pipeline");

	if (!(priv->pipeline)) {
		g_set_error (error,
		             error_quark (),
		             INITIALIZATION_ERROR,
		             "Failed to create pipeline element");

		return FALSE;
	}

	priv->source       = gst_element_factory_make ("filesrc", "source");
	priv->decodebin    = gst_element_factory_make ("decodebin2", "decodebin2");

	if (!(priv->source &&
	      priv->decodebin)) {
			  g_set_error (error,
			               error_quark (),
			               INITIALIZATION_ERROR,
			               "Failed to create source and decodebin elements");
			  gst_object_unref (priv->pipeline);
			  if (priv->source)
				  gst_object_unref (priv->source);
			  if (priv->decodebin)
				  gst_object_unref (priv->decodebin);
			  return FALSE;
		  }

	gst_bin_add_many (GST_BIN(priv->pipeline),
	                  priv->source,
	                  priv->decodebin,
	                  NULL);

	if (!gst_element_link_many(priv->source, priv->decodebin, NULL)) {
		g_set_error (error,
		             error_quark (),
		             INITIALIZATION_ERROR,
		             "Failed to link source and decoding components");
		gst_object_unref (priv->pipeline);
		return FALSE;
	}

	priv->sinkbin      = gst_bin_new("sink bin");
	priv->video_scaler = gst_element_factory_make ("videoscale", "video_scaler");
	priv->video_color  = gst_element_factory_make ("ffmpegcolorspace", "video_color");
	priv->video_filter = gst_element_factory_make ("capsfilter", "video_filter");
	priv->video_sink   = gst_element_factory_make ("fakesink", "video_sink");

	if (!(priv->sinkbin &&
	      priv->video_scaler &&
	      priv->video_color &&
	      priv->video_filter &&
	      priv->video_sink)) {
			  g_set_error (error,
			               error_quark (),
			               INITIALIZATION_ERROR,
			               "Failed to create scaler and sink elements");
			  gst_object_unref (priv->pipeline);
			  if (priv->sinkbin)
				  gst_object_unref (priv->sinkbin);
			  if (priv->video_scaler)
				  gst_object_unref (priv->video_scaler);
			  if (priv->video_filter)
				  gst_object_unref (priv->video_filter);
			  if (priv->video_color)
				  gst_object_unref (priv->video_color);
			  if (priv->video_sink)
				  gst_object_unref (priv->video_sink);
			  return FALSE;
		  }

	gst_bin_add_many (GST_BIN(priv->sinkbin),
	                  priv->video_scaler,
	                  priv->video_color,
	                  priv->video_filter,
	                  priv->video_sink,
	                  NULL);

	if (!gst_element_link_many(priv->video_scaler, priv->video_color, priv->video_filter,  NULL)) {
		g_set_error (error,
		             error_quark (),
		             INITIALIZATION_ERROR,
		             "Failed to link scaler, color, filter and encoder elements");
		gst_object_unref (priv->pipeline);
		gst_object_unref (priv->sinkbin);
		return FALSE;
	}

	caps = gst_caps_new_simple ("video/x-raw-rgb",
	                            "bpp", G_TYPE_INT, 24,
	                            "depth", G_TYPE_INT, 24,
	                            NULL);

	if (!gst_element_link_filtered(priv->video_filter, priv->video_sink, caps)) {
		g_set_error (error,
		             error_quark (),
		             INITIALIZATION_ERROR,
		             "Failed to link filter and encoder elements with caps");
		gst_object_unref (priv->pipeline);
		gst_object_unref (priv->sinkbin);
		gst_caps_unref (caps);
		return FALSE;
	}

	gst_caps_unref (caps);

	videopad = gst_element_get_pad (priv->video_scaler, "sink");
	gst_element_add_pad (priv->sinkbin, gst_ghost_pad_new ("sink", videopad));
	gst_object_unref (videopad);

	gst_bin_add (GST_BIN (priv->pipeline), priv->sinkbin);

	/* Set videoscale to fastest resampling mode (nearest neighbour) */
	g_object_set (priv->video_scaler, "method", 0, NULL);

	g_object_set (priv->video_sink, "signal-handoffs", TRUE, NULL);

	/* Connect signal for new pads */
	g_signal_connect (priv->decodebin, "new-decoded-pad",
	                  G_CALLBACK (newpad_callback), pipe);


	/* Connect signal for analysing new streams (we only care about video) */
	g_signal_connect (priv->decodebin, "autoplug-continue",
	                  G_CALLBACK (stream_continue_callback), pipe);


	/* Connect signal for handoffs */
	g_signal_connect (priv->video_sink, "handoff",
	                  G_CALLBACK (buffer_callback), pipe);

	return TRUE;
}

static void
set_caps_filter (ThumberPipe *pipe, GstCaps *caps)
{
	int           i;
        guint         flav_h, flav_w, flav_min_h, flav_min_w;
	guint         height, width, min_height, min_width;
	ThumberPipePrivate *priv;
	
	priv = THUMBER_PIPE_GET_PRIVATE (pipe);
	
	thumber_flavor_get_size (priv->flavor, &flav_w, &flav_h, &flav_min_w, &flav_min_h);

        /* If the image is rotated we need to switch height <-> width and then turn
         * when creating the actual thumbnail.
         */

	if (strcmp(thumber_flavor_get_rotation (priv->flavor), "Rotate") == 0) {
                if (priv->rotation == ROTATE90 || priv->rotation == ROTATE270) {
                        /* If height <-> width rotated */
                        width = flav_h;
                        height = flav_w;
                        min_width = flav_min_h;
                        min_height = flav_min_w;                
                } else {
                        /* Normal or upside down */
                        width = flav_w;
                        height = flav_h;
                        min_width = flav_min_w;
                        min_height = flav_min_h;
                }
        } else {
                width = flav_w;
                height = flav_h;
                min_width = flav_min_w;
                min_height = flav_min_h;      
        }
	
	/* We are setting the capsfilter here in order to force original aspect-ratio.
	   This is a bit dirty but the best we can do without writing a component to
	   do some negotiations (like ximagesink for instance).
	*/
	
	for (i = caps->structs->len - 1; i >= 0; i--) {
		gint w, h;
		int o;
		GstStructure *str;
		
		str = gst_caps_get_structure (caps, i);
		
		if (gst_structure_get_int (str, "width", &w) &&
		    gst_structure_get_int (str, "height", &h) && w && h) {
			double width_scale;
			double height_scale;
			guint adj_width;
			guint adj_height;
			
			width_scale = ((double)width)/(double)w;
			height_scale = ((double)height)/(double)h;
			
			if (width_scale < height_scale) {
				adj_width = width;
				adj_height = (int)(h*width_scale);
			} else {
				adj_width = (int)(w*height_scale);
				adj_height = height;
			}
			
			if (min_width) {
				if (adj_width < min_width) {
					width_scale = ((double)min_width)/(double)adj_width;
					adj_height = adj_height * width_scale;
					adj_width = min_width;
				}
			}
			
			if (min_height) {
				if (adj_height < min_height) {
					height_scale = ((double)min_height)/(double)adj_height;
					adj_width = adj_width * height_scale;
					adj_height = min_height;
				}
			}
			
			/* If the normal rules would lead to any upscaling, go with the original */
			if ((adj_height > h) || (adj_width > w)) {
				adj_width = w;
				adj_height = h;
			}
			
			caps = gst_caps_new_simple ("video/x-raw-rgb",
						    "width", G_TYPE_INT, adj_width,
						    "height", G_TYPE_INT, adj_height,
						    NULL);
			
			g_object_set (G_OBJECT (priv->video_filter),
				      "caps", caps,
				      NULL);
			
                        priv->width = adj_width;
                        priv->height = adj_height;

			/* We are done, found one */
			return;
		}
	}
}

static void
pad_notify_caps_callback (GstPad *pad, GParamSpec *pspec, gpointer *user_data)
{
	GstCaps *caps;
	ThumberPipe *pipe;
	
	pipe = user_data;
	
	g_object_get (pad, "caps", &caps, NULL);
	
	if (caps == NULL) {
		return;
	}
	
	set_caps_filter (pipe, caps);
}

static void
newpad_callback (GstElement       *decodebin,
                 GstPad           *pad,
                 gboolean          last,
                 ThumberPipe      *pipe)
{
	ThumberPipePrivate *priv;
	
	GstCaps      *caps;
	GstStructure *str;
	GstPad       *videopad;

	g_return_if_fail (decodebin != NULL);
	g_return_if_fail (pad != NULL);
	g_return_if_fail (pipe != NULL);
	
	priv = THUMBER_PIPE_GET_PRIVATE (pipe);
	
	videopad = gst_element_get_static_pad (priv->sinkbin, "sink");
	
	if (!videopad) {
		return;
	}

	if (GST_PAD_IS_LINKED (videopad)) {
		gst_object_unref (videopad);
		return;
	}
	
	caps = gst_pad_get_caps (pad);
	str  = gst_caps_get_structure (caps, 0);

	g_signal_connect (pad, "notify::caps",
			  G_CALLBACK (pad_notify_caps_callback), pipe);

	if (!g_strrstr (gst_structure_get_name (str), "video")) {
		gst_object_unref (videopad);
		gst_caps_unref (caps);
		return;
	}
	gst_caps_unref (caps);
	gst_pad_link (pad, videopad);

	gst_object_unref (videopad);
}

static gboolean
stream_continue_callback (GstElement    *bin,
                          GstPad        *pad,
                          GstCaps       *caps,
                          ThumberPipe   *pipe)
{
	int           i;
	ThumberPipePrivate *priv;

	priv = THUMBER_PIPE_GET_PRIVATE (pipe);

	g_return_val_if_fail (bin != NULL, FALSE);
	g_return_val_if_fail (pad != NULL, FALSE);
	g_return_val_if_fail (caps != NULL, FALSE);
	g_return_val_if_fail (pipe != NULL, FALSE);

	for (i = caps->structs->len - 1; i >= 0; i--) {
		GstStructure *str;

		str = gst_caps_get_structure (caps, i);

		/* Because of some inconsistencies (audio/?? container
		   containing video) we have a blacklist here. */

		if (strcasecmp (gst_structure_get_name (str), "audio/mpeg") == 0 ||
                    strcasecmp (gst_structure_get_name (str), "audio/ac3" ) == 0 ||
                    strcasecmp (gst_structure_get_name (str), "audio/x-ac3" ) == 0 ||
                    strcasecmp (gst_structure_get_name (str), "audio/x-private1-ac3" ) == 0 ||
                    strcasecmp (gst_structure_get_name (str), "audio/x-eac3" ) == 0 ||
                    strcasecmp (gst_structure_get_name (str), "audio/x-gst-fourcc-ec-3") == 0 ||
		    strcasecmp (gst_structure_get_name (str), "audio/amr") == 0 ||
		    strcasecmp (gst_structure_get_name (str), "audio/amr-wb") == 0 ||
		    strcasecmp (gst_structure_get_name (str), "audio/x-pn-realaudio") == 0) {
		        /* Do not plug these at all */
		        return FALSE;
		}
	}

	/*	set_caps_filter (pipe, caps);*/

	return TRUE;
}

static void
buffer_callback (GstElement    *bin,
                 GstBuffer     *buffer,
                 GstPad        *pad,
                 ThumberPipe   *pipe)
{
	GstMessage         *msg  = NULL;
	GstStructure       *str  = NULL;
	GstBus             *bus  = NULL;
	unsigned char      *data = NULL;
	gint                size;
	ThumberPipePrivate *priv;

	priv = THUMBER_PIPE_GET_PRIVATE (pipe);
	data = (unsigned char *) GST_BUFFER_DATA (buffer);
	size = GST_BUFFER_SIZE (buffer);

	if (!check_for_valid_thumb (data,
				    size)) {
		priv->rejected_frame_count++;

		if (priv->rejected_frame_count < FRAME_REJECT_LIMIT) {
		        return;
		} else {
		        /*g_debug ("Reached limit for rejected frames, accepting");*/
		}
	}

	if (create_thumbnail (data, priv->uri, priv->width, priv->height, priv->rotation, priv->flavor)) {
	        str = gst_structure_new ("frame-encoded", NULL);
		msg = gst_message_new_application (GST_OBJECT (bin), str);
		bus = gst_element_get_bus (priv->pipeline);
		gst_bus_post ( bus, msg);
	} else {
	        g_warning ("Failed to create the thumbnail for %s", priv->uri);
	}
}

static gboolean
wait_for_state_change (ThumberPipe *pipe,
                       GstState     state,
                       GError     **error)
{
	ThumberPipePrivate *priv;
	GstBus             *bus;
	gint64              timeout = PIPE_TIMEOUT * GST_SECOND;

	priv = THUMBER_PIPE_GET_PRIVATE (pipe);

	bus = gst_element_get_bus (priv->pipeline);

	while (TRUE) {
		GstMessage *message;
		GstElement *src;
                
		message = gst_bus_timed_pop (bus, timeout);
                
		if (!message) {
			g_set_error (error,
			             error_quark (),
			             RUNNING_ERROR,
			             "Pipeline timed out");
			goto error;
		}

		src = (GstElement*)GST_MESSAGE_SRC (message);

		switch (GST_MESSAGE_TYPE (message)) {
                case GST_MESSAGE_STATE_CHANGED: {
                        GstState old, new, pending;
                        
                        if (src == priv->pipeline) {
                                
                                gst_message_parse_state_changed (message, &old, &new, &pending);
                                
                                if (new == state) {
                                        gst_message_unref (message);
                                        goto success;
                                }
                        }
                        break;
                }
                case GST_MESSAGE_ERROR: {
                        GError *lerror = NULL;
                        
                        gst_message_parse_error (message, &lerror, NULL);
                        
                        if (lerror != NULL) {
                                g_propagate_error (error, lerror);
                        } else {
                                g_set_error (error,
                                             error_quark (),
                                             RUNNING_ERROR,
                                             "Undefined error running the pipeline");
                        }
                        
                        gst_message_unref (message);
                        goto error;
                        
                        break;
                }
                case GST_MESSAGE_EOS: {
                        g_set_error (error,
                                     error_quark (),
                                     RUNNING_ERROR,
                                     "Reached end-of-file without proper content");
                        
                        gst_message_unref (message);
                        goto error;
                        
                        break;
                }
                case GST_MESSAGE_TAG: {
                        GstTagList         *tags;
                        gchar              *value = NULL;
                        gst_message_parse_tag (message, &tags);
                        
                        gst_tag_list_get_string (tags,GST_TAG_IMAGE_ORIENTATION,&value);
                        if (value) {
                                if ( strcmp(value, "rotate-90") == 0) {
                                        priv->rotation = ROTATE90;
                                } else if ( strcmp(value, "rotate-180") == 0) {
                                        priv->rotation = ROTATE180;
                                } else if ( strcmp(value, "rotate-270") == 0) {
                                        priv->rotation = ROTATE270;
                                }
                        }
                        
                }        
                default:
                        /* Nothing to do here */
                        break;
		}
                
		gst_message_unref (message);
	}

	g_assert_not_reached ();

	error:
		gst_object_unref (bus);
		return FALSE;

	success:
		gst_object_unref (bus);
		return TRUE;
}

static gboolean
wait_for_image_buffer (ThumberPipe *pipe,
                       const gchar *uri,
                       GError     **error)
{
	ThumberPipePrivate *priv;
	GstBus             *bus;
	gint64              timeout = PIPE_TIMEOUT * GST_SECOND;

	priv = THUMBER_PIPE_GET_PRIVATE (pipe);

	bus = gst_element_get_bus (priv->pipeline);

	while (TRUE) {
		GstMessage *message;
		GstElement *src;

		message = gst_bus_timed_pop (bus, timeout);

		if (!message) {
			g_set_error (error,
			             error_quark (),
			             RUNNING_ERROR,
			             "Pipeline timed out getting the buffer");
			goto error;
		}

		src = (GstElement*)GST_MESSAGE_SRC (message);

		switch (GST_MESSAGE_TYPE (message)) {
			case GST_MESSAGE_ERROR: {
				GError *lerror = NULL;

				gst_message_parse_error (message, &lerror, NULL);

				if (lerror != NULL) {
					g_propagate_error (error, lerror);
				} else {
					g_set_error (error,
					             error_quark (),
					             RUNNING_ERROR,
					             "Undefined error running the pipeline");
				}

				gst_message_unref (message);
				goto error;
				break;
			}
			case GST_MESSAGE_EOS: {
				g_set_error (error,
				             error_quark (),
				             RUNNING_ERROR,
				             "Reached end-of-file without proper content");
				gst_message_unref (message);
				goto error;
				break;
			}
			case GST_MESSAGE_APPLICATION: {
				const GstStructure *s = gst_message_get_structure (message);
				const gchar *name = gst_structure_get_name (s);

				if (strcmp (name, "frame-encoded") == 0) {
					goto success;
				}
				break;
			}
			default:
				/* Nothing to do here */
				break;
		}

		gst_message_unref (message);
	}

	g_assert_not_reached ();

	error:
		gst_object_unref (bus);
		return FALSE;

	success:
		gst_object_unref (bus);
		return TRUE;
}

static gboolean
check_for_valid_thumb (unsigned char *buffer,
		       const gint size)
{
       /* guchar* pixels = gdk_pixbuf_get_pixels(pixbuf); */
       /* int     rowstride = gdk_pixbuf_get_rowstride(pixbuf); */
       /* int     height = gdk_pixbuf_get_height(pixbuf); */
       /* int     samples = (rowstride * height); */

       int samples = size;

       int     i;
       float   avg = 0.0f;
       float   variance = 0.0f;

       g_return_val_if_fail ( size > 0, FALSE);

       /* We calculate the variance of one element (bpp=24) */
 
       for (i = 0; i < samples; i=i+3) {
               avg += (float) buffer[i];
       }
       avg /= ((float) samples);

       /* Calculate the variance */
       for (i = 0; i < samples; i=i+3) {
               float tmp = ((float) buffer[i] - avg);
              variance += tmp * tmp;
       }
       variance /= ((float) (samples - 1));

       return (variance > VALID_VARIANCE_THRESHOLD);
 }


static gchar *
compute_checksum (GChecksumType  checksum_type,
                  const guchar  *data,
                  gsize          length)
{
	GChecksum *checksum;
	gchar *retval;

	checksum = g_checksum_new (checksum_type);
	if (!checksum)
		return NULL;

	g_checksum_update (checksum, data, length);
	retval = g_strdup (g_checksum_get_string (checksum));
	g_checksum_free (checksum);

	return retval;
}

static gchar *
filename_for_thumb (const gchar *uri, const gchar *flavor)
{
	gchar *name;
	gchar *dir;
	gchar *checksum;
	gchar *filename;

	dir = g_build_filename (g_get_home_dir (), ".thumbnails", flavor, NULL);

	if(!g_file_test (dir, G_FILE_TEST_EXISTS)) {
		g_mkdir_with_parents (dir, 0770);
	}

	checksum = compute_checksum (G_CHECKSUM_MD5, (const guchar *)uri, strlen (uri));
	name = g_strdup_printf ("%s.jpeg", checksum);

	filename = g_build_filename (dir, name, NULL);

	g_free (checksum);
	g_free (dir);
	g_free (name);

	return filename;
}

static gboolean
write_thumbnail_info (const gchar  *filename,
                      gchar        *uri,
                      guint64       mtime,
                      GCancellable *cancellable,
                      GError      **error)
{
	struct utimbuf  buf;
	gchar          *path;
	const gchar    *home;
	FILE           *journal;
	size_t          len;
	gsize           max_pos;
	gsize           next_len;
	guint           s_len;
	gchar          *cur;
	GMappedFile    *mapped;

	buf.actime = buf.modtime = mtime;
	utime (filename, &buf);

	home = g_getenv ("HOME") != NULL ? g_getenv ("HOME") : g_get_home_dir ();
	path = g_build_filename (home, ".thumbnails", ".data", NULL);

	mapped = g_mapped_file_new (path, FALSE, NULL);

	if (mapped)
	{
		cur = g_mapped_file_get_contents (mapped);
		max_pos = (gsize) (cur + g_mapped_file_get_length (mapped));
		s_len = strlen (filename);

		while (((gsize)cur) < max_pos) 
		{

			if ((((gsize)cur) + s_len) > max_pos)
				break;

			next_len = strnlen (cur, max_pos - ((gsize)cur)) + 2;

			if (g_strrstr_len (cur, s_len, filename) && 
			    ((gsize)cur) + s_len + 3 < max_pos   &&
			    (cur[s_len + 0] == ' ')              && 
			    (cur[s_len + 1] == '=')              && 
			    (cur[s_len + 2] == ' '))
			{
				g_free (path);
				g_mapped_file_free (mapped);
				return TRUE;
			}
			cur += next_len;
		}
		g_mapped_file_free (mapped);
	}

	journal = fopen (path, "a");
	g_free (path);

	if (journal) 
	{
		flock(fileno (journal), LOCK_EX);
		path = g_strdup_printf ("%s = %s", filename, uri);
		len = strlen (path);
		write (fileno (journal), path, len);
		write (fileno (journal), "\0\0", 2);
		fsync (fileno (journal));
		g_free (path);
		flock(fileno (journal), LOCK_UN);
		fclose (journal);
	}

	return TRUE;
}

static void
jpeg_error_exit (j_common_ptr cinfo)
{
        struct tej_error_mgr *h = (struct tej_error_mgr *) cinfo->err;
        (*cinfo->err->output_message)(cinfo);
        longjmp (h->setjmp_buffer, 1);
}

static gboolean
create_thumbnail (unsigned char       *buffer,
                  const gchar         *uri,
                  const guint          buffer_width,
                  const guint          buffer_height,
                  const guint          rotation,
                  const ThumberFlavor *flavor)
{
	FILE                       *out;
	unsigned char              *data;
	struct jpeg_compress_struct comp;
	struct tej_error_mgr        error;
	gchar                      *filename;
	gchar                      *filename_tmp;
	gint                        i,j,k;
	const gchar                *flavor_str = thumber_flavor_get_name (flavor);
	gint                        row_stride;
	JSAMPROW                    row_pointer[1];
	gint                        width, height, min_width, min_height;
	gint                        flavor_width, flavor_height, flavor_min_width, flavor_min_height;
	gint                        crop_offset;
        gint                        data_width;
        gint                        data_height;
	gint                        end_width;
	gint                        end_height;
	gint                        crop_width;
	gint                        crop_height;
	struct utimbuf              buf;
	GFile                      *orig_file;
	GFileInfo                  *file_info;
	guint64                     mtime;

	orig_file = g_file_new_for_uri (uri);
	file_info = g_file_query_info (orig_file, G_FILE_ATTRIBUTE_TIME_MODIFIED ","
	                               G_FILE_ATTRIBUTE_TIME_ACCESS ","
	                               G_FILE_ATTRIBUTE_TIME_CHANGED ","
	                               G_FILE_ATTRIBUTE_TIME_CREATED, 0, NULL, NULL);

	mtime = g_file_info_get_attribute_uint64 (file_info, G_FILE_ATTRIBUTE_TIME_MODIFIED);

	g_object_unref (orig_file);
	g_object_unref (file_info);

	filename = filename_for_thumb (uri, flavor_str);
	filename_tmp = g_strdup_printf ("%s.tmp", filename);

	out = fopen (filename_tmp, "wb");

	if (!out) {
		g_warning ("Opening the thumbnail output file failed");
		g_free (filename_tmp);
		g_free (filename);
		return FALSE;
	}

	thumber_flavor_get_size (flavor, &flavor_width, &flavor_height, &flavor_min_width, &flavor_min_height);

        /* FIXME remove the allocation if not needed */
        data = g_malloc(buffer_width * buffer_height * 3);

	if (strcmp(thumber_flavor_get_rotation (flavor), "Rotate") == 0) {
                
                width = flavor_width;
                height = flavor_height;
                min_width = flavor_min_width;
                min_height = flavor_min_height;

                /* If rotated, rotate the data. FIXME There needs to be something more efficient we can do */
                switch(rotation) {
                default:
                case ROTATE0:
                        data_width  = buffer_width;
                        data_height = buffer_height;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[i*buffer_width*3 + j*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                case ROTATE90:
                        data_width  = buffer_height;
                        data_height = buffer_width;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[j*data_width*3 + (buffer_height-i-1)*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                case ROTATE180:
                        data_width  = buffer_width;
                        data_height = buffer_height;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[(buffer_height-i-1)*buffer_width*3 + (buffer_width-j-1)*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                case ROTATE270:
                        data_width  = buffer_height;
                        data_height = buffer_width;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[(buffer_width-j-1)*data_width*3 + i*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                }
        } else if (strcmp(thumber_flavor_get_rotation (flavor), "Transpose") == 0) {
                
                /* If rotated, rotate the data. FIXME There needs to be something more efficient we can do */
                switch(rotation) {
                default:
                case ROTATE0:
                        width = flavor_width;
                        height = flavor_height;
                        min_width = flavor_min_width;
                        min_height = flavor_min_height;
                        
                        data_width  = buffer_width;
                        data_height = buffer_height;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[i*buffer_width*3 + j*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                case ROTATE90:
                        width = flavor_height;
                        height = flavor_width;
                        min_width = flavor_min_height;
                        min_height = flavor_min_width;

                        data_width  = buffer_height;
                        data_height = buffer_width;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[j*data_width*3 + (buffer_height-i-1)*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                case ROTATE180:
                        width = flavor_width;
                        height = flavor_height;
                        min_width = flavor_min_width;
                        min_height = flavor_min_height;

                        data_width  = buffer_width;
                        data_height = buffer_height;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[(buffer_height-i-1)*buffer_width*3 + (buffer_width-j-1)*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                case ROTATE270:
                        width = flavor_height;
                        height = flavor_width;
                        min_width = flavor_min_height;
                        min_height = flavor_min_width;

                        data_width  = buffer_height;
                        data_height = buffer_width;
                        
                        row_stride = buffer_width * 3;
                        
                        if (row_stride % 4) {
                                row_stride += (4 - row_stride % 4);
                        }
                        
                        for (i=0;i<buffer_height;i++)
                                for (j=0;j<buffer_width;j++)
                                        for (k=0;k<3;k++)
                                                data[(buffer_width-j-1)*data_width*3 + i*3 + k] = buffer[i*row_stride + j*3 + k];
                        break;
                }
        } else {
                width = flavor_width;
                height = flavor_height;
                min_width = flavor_min_width;
                min_height = flavor_min_height;

                data_width  = buffer_width;
                data_height = buffer_height;
                        
                row_stride = buffer_width * 3;
                        
                if (row_stride % 4) {
                        row_stride += (4 - row_stride % 4);
                }
                for (i=0;i<buffer_height;i++)
                        for (j=0;j<buffer_width;j++)
                                for (k=0;k<3;k++)
                                        data[i*buffer_width*3 + j*3 + k] = buffer[i*row_stride + j*3 + k];
        }

        end_width = MIN (width, data_width);
        end_height = MIN (height, data_height);
        
        if (width < data_width) {
                crop_width = (data_width - width)/2;
        } else {
                crop_width = 0;
        }
        
        if (height < data_height) {
                crop_height = (data_height - height)/2;
        } else {
                crop_height = 0;
        }
        
	comp.err = jpeg_std_error (&error.jpeg);
	error.jpeg.error_exit = jpeg_error_exit;
	if (setjmp (error.setjmp_buffer)) {
	        goto fail;
	}
	jpeg_create_compress (&comp);
	jpeg_stdio_dest (&comp, out);
	comp.image_width = end_width;
	comp.image_height = end_height;
	comp.input_components = 3;
	comp.in_color_space = JCS_RGB;
	jpeg_set_defaults (&comp);
	jpeg_set_quality (&comp, 90, TRUE);
	jpeg_start_compress (&comp, TRUE);

        /* row_stride = data_width * comp.input_components; */
	/* /\* Padded to full words *\/ */
	/* if (row_stride % 4) { */
        /*         row_stride += (4 - row_stride % 4); */
	/* } */

	while (comp.next_scanline<comp.image_height) {
		row_pointer[0]=&data[(comp.next_scanline+crop_height)*data_width*comp.input_components + crop_width*comp.input_components];
		jpeg_write_scanlines(&comp,row_pointer,1);
	}

	jpeg_finish_compress (&comp);
	fflush (out);
	fclose (out);
	jpeg_destroy_compress (&comp);

	buf.actime = buf.modtime = mtime;
	utime (filename_tmp, &buf);

	g_rename (filename_tmp, filename);

	write_thumbnail_info (filename, uri, mtime, NULL, NULL);

        g_free (data);

	g_free (filename_tmp);
	g_free (filename);

	return TRUE;
 fail:
        g_free (data);

	g_free (filename_tmp);
	g_free (filename);

	return FALSE;
}


