/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/*
 * This file is part of maemo-video-thumbnailer package
 * A plugin for Tumbler to create video thumbnails
 *
 * Copyright (C) 2009 Nokia Corporation.  All Rights reserved.
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define THUMBER_THUMBER_ERROR_DOMAIN "ThumberThumberError"
#define DEFAULT_QUIT_TIMEOUT 30

#include "gst-thumb-thumber.h"

#include <glib.h>
#include <gio/gio.h>
#include <gst/gst.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>

#ifdef HAVE_RESOURCE
#include <resource.h>
#endif

#include "gst-video-thumbnailer-marshal.h"
#include "gst-thumb-main.h"
#include "gst-thumb-pipe.h"
#include "gst-thumb-flavor.h"

enum ThumberState {
	THUMBER_STATE_NULL,
	THUMBER_STATE_PENDING_STOP,
	THUMBER_STATE_STOPPED,
	THUMBER_STATE_PAUSED,
	THUMBER_STATE_WORKING,
};
typedef enum ThumberState ThumberState;

enum {
	NO_ERROR,
	INITIALIZATION_ERROR,
	RUNNING_ERROR,
	DBUS_ERROR
};

void thumber_dbus_method_queue   (Thumber *object, gchar *uri, gchar *mime_hint,
                                  gchar *flavor, gboolean priority, DBusGMethodInvocation *context);
void thumber_dbus_method_dequeue (Thumber *object, guint handle, DBusGMethodInvocation *context);


#define gst_video_thumbnailer_queue   thumber_dbus_method_queue
#define gst_video_thumbnailer_dequeue thumber_dbus_method_dequeue

#include "gst-video-thumbnailer-glue.h"

static gboolean thumber_process_func (gpointer data);
static void     thumber_set_state (Thumber *thumber, ThumberState state);

static void     request_resources (Thumber *thumber);
static void     release_resources (Thumber *thumber);

static void     load_flavors (Thumber *thumber);
const ThumberFlavor * thumber_get_flavor (Thumber     *thumber,
                                          const gchar *flavor);
G_DEFINE_TYPE (Thumber, thumber, G_TYPE_OBJECT)

typedef struct FileInfo FileInfo;
typedef struct TaskInfo TaskInfo;

typedef struct {
	/* Properties */
	guint            idle_id;
	guint            quit_timeout_id;
	gint             quit_timeout;

	TaskInfo        *current_task;

	GQueue          *task_queue;
	GQueue          *file_queue;

	GList           *flavors;

	ThumberPipe     *pipe;

	ThumberState     state;
#ifdef HAVE_RESOURCE
	resource_set_t  *resource;
#endif

} ThumberPrivate;

struct FileInfo {
	gchar               *uri;
	guint                id;
	const ThumberFlavor *flavor;
};

struct TaskInfo {
	guint   id;
	gchar  *mime;
	GSList *files;
};

enum {
	STARTED_SIGNAL,
	FINISHED_SIGNAL,
	READY_SIGNAL,
	ERROR_SIGNAL,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0, };

static GQuark
error_quark (void)
{
	return g_quark_from_static_string (THUMBER_THUMBER_ERROR_DOMAIN);
}

FileInfo *
file_info_new (const gchar *uri, const ThumberFlavor *flavor, guint id)
{
	FileInfo *info;

	info            = g_slice_new0 (FileInfo);
	info->uri       = g_strdup (uri);
	info->flavor    = flavor;
	info->id        = id;
	return info;
}

void
file_info_free (FileInfo *info)
{
	if (info->uri) {
		g_free (info->uri);
	}

	g_slice_free (FileInfo, info);
}

TaskInfo *
task_info_new (const gchar *mime)
{
	TaskInfo     *info;
	static guint  id_counter = 0;

	info = g_slice_new0 (TaskInfo);
	info->id = ++id_counter;
	info->mime  = g_strdup(mime);
	info->files = NULL;

	return info;
}

void
task_info_add_uri (TaskInfo *info, gchar *uri, const ThumberFlavor *flavor)
{
	FileInfo *file;
	file = file_info_new (uri, flavor, info->id);
	info->files = g_slist_prepend (info->files, file);
}

void
task_info_add_uris (TaskInfo *info, GStrv uris, const ThumberFlavor *flavor)
{
	guint i;

	for (i=0;uris[i];i++) {
		task_info_add_uri (info, uris[i], flavor);
	}
}

void
task_info_free (TaskInfo *info, gboolean free_content)
{
	if (info->mime) {
		g_free (info->mime);
	}

	if (free_content) {
		g_slist_foreach (info->files,
		                 (GFunc)file_info_free,
		                 NULL);
	}

	g_slist_free (info->files);

	g_slice_free (TaskInfo, info);
}

enum {
	PROP_0,
	PROP_TIMEOUT
};

#define THUMBER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj),TYPE_THUMBER, ThumberPrivate))

Thumber *
thumber_new ()
{
	return g_object_new (TYPE_THUMBER, NULL);
}

static gint
thumber_get_timeout (Thumber *thumber)
{
	ThumberPrivate *priv;

	priv = THUMBER_GET_PRIVATE (thumber);
	return priv->quit_timeout;
}

static void
thumber_set_timeout (Thumber *thumber, gint timeout)
{
	ThumberPrivate *priv;

	priv = THUMBER_GET_PRIVATE (thumber);
	priv->quit_timeout = timeout;
}

static void
thumber_set_property (GObject      *object,
                      guint         prop_id,
                      const GValue *value,
                      GParamSpec   *pspec)
{
	switch (prop_id) {
		case PROP_TIMEOUT:
			thumber_set_timeout (THUMBER (object), g_value_get_int (value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}


static void
thumber_get_property (GObject    *object,
                      guint       prop_id,
                      GValue     *value,
                      GParamSpec *pspec)
{
	ThumberPrivate *priv;

	priv = THUMBER_GET_PRIVATE (object);

	switch (prop_id) {
		case PROP_TIMEOUT:
			g_value_set_int (value,
			                 thumber_get_timeout (THUMBER (object)));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
	}
}

static void
thumber_finalize (GObject *object)
{
	ThumberPrivate *priv;

	priv = THUMBER_GET_PRIVATE (object);

#ifdef HAVE_RESOURCE
	resource_set_destroy (priv->resource);
	priv->resource = NULL;
#endif
	if (priv->pipe) {
		g_object_unref (priv->pipe);
	}

	g_list_foreach (priv->flavors, (GFunc) g_object_unref, NULL);
	g_list_free (priv->flavors);

	g_queue_foreach (priv->task_queue, (GFunc) task_info_free, (gpointer)TRUE);
	g_queue_free (priv->task_queue);

	g_queue_foreach (priv->file_queue, (GFunc) file_info_free, NULL);
	g_queue_free (priv->file_queue);

	if (priv->idle_id) {
		g_source_remove (priv->idle_id);
	}

	gst_deinit ();

	G_OBJECT_CLASS (thumber_parent_class)->finalize (object);
}

static void
thumber_class_init (ThumberClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	object_class->finalize     = thumber_finalize;
	object_class->set_property = thumber_set_property;
	object_class->get_property = thumber_get_property;

	g_object_class_install_property (object_class,
	                                 PROP_TIMEOUT,
	                                 g_param_spec_int ("timeout",
	                                                   "Timeout",
	                                                   "Time after which we quit if nothing to do",
	                                                   -1,
	                                                   3600,
	                                                   DEFAULT_QUIT_TIMEOUT,
	                                                   G_PARAM_READWRITE));


	signals[READY_SIGNAL] =
		g_signal_new ("ready",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (ThumberClass, ready),
		              NULL, NULL,
		              gst_video_thumbnailer_marshal_VOID__UINT_STRING,
		              G_TYPE_NONE,
		              2,
		              G_TYPE_UINT,
		              G_TYPE_STRING);

	signals[STARTED_SIGNAL] =
		g_signal_new ("started",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (ThumberClass, started),
		              NULL, NULL,
		              g_cclosure_marshal_VOID__UINT,
		              G_TYPE_NONE,
		              1,
		              G_TYPE_UINT);

	signals[FINISHED_SIGNAL] =
		g_signal_new ("finished",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (ThumberClass, finished),
		              NULL, NULL,
		              g_cclosure_marshal_VOID__UINT,
		              G_TYPE_NONE,
		              1,
		              G_TYPE_UINT);

	signals[ERROR_SIGNAL] =
		g_signal_new ("error",
		              G_OBJECT_CLASS_TYPE (object_class),
		              G_SIGNAL_RUN_LAST,
		              G_STRUCT_OFFSET (ThumberClass, error),
		              NULL, NULL,
		              gst_video_thumbnailer_marshal_VOID__UINT_STRING_INT_STRING,
		              G_TYPE_NONE,
		              4,
		              G_TYPE_UINT,
		              G_TYPE_STRING,
		              G_TYPE_INT,
		              G_TYPE_STRING);

	g_type_class_add_private (object_class, sizeof (ThumberPrivate));
}

static void
thumber_init (Thumber *object)
{
	GError *error = NULL;
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (object);

	gst_init (NULL, NULL);

	priv->task_queue = g_queue_new ();
	priv->file_queue = g_queue_new ();

	priv->idle_id    = 0;
	priv->quit_timeout_id = 0;

	priv->current_task = NULL;

	priv->pipe = thumber_pipe_new ();

	if (!thumber_pipe_initialize (priv->pipe, &error)) {
		g_error ("Problem initializing the pipeline: %s",
		         error->message);
	}

	priv->state = THUMBER_STATE_STOPPED;

	load_flavors (object);
}

#ifdef HAVE_RESOURCE
static void
resource_grant_handler (resource_set_t *resource_set,
                        uint32_t        resources,
                        void           *data)
{
	Thumber        *thumber;
	ThumberPrivate *priv;

	thumber = data;
	priv = THUMBER_GET_PRIVATE (thumber);

	if (resources && RESOURCE_VIDEO_PLAYBACK) {
		if (priv->state == THUMBER_STATE_PAUSED) {
			thumber_set_state (thumber,THUMBER_STATE_WORKING);
		} else {
			g_warning ("Policy is granting resources without a request");
		}
	} else {
		if (priv->state == THUMBER_STATE_WORKING) {
                        thumber_set_state (thumber,THUMBER_STATE_PAUSED);
			release_resources (thumber);			
		} else if (priv->state == THUMBER_STATE_PENDING_STOP) {
			thumber_set_state (thumber,THUMBER_STATE_STOPPED);
		}
	}

}
#endif

#ifdef HAVE_RESOURCE
static void
resource_advice_handler (resource_set_t *resource_set,
                         uint32_t        resources,
                         void           *data)
{
	Thumber        *thumber;
	ThumberPrivate *priv;

	thumber = data;
	priv = THUMBER_GET_PRIVATE (thumber);

	switch (priv->state) {
		case THUMBER_STATE_PENDING_STOP:
			if (!(resources && RESOURCE_VIDEO_PLAYBACK)) {
				release_resources (thumber);
			}
			break;
		case THUMBER_STATE_PAUSED:
			if (resources && RESOURCE_VIDEO_PLAYBACK) {
				request_resources (thumber);
			}
			break;
		default:
			/* Nothing */
			break;
	}
}
#endif

static void
request_resources (Thumber *thumber)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (thumber);

	if (priv->state != THUMBER_STATE_WORKING) {
#ifdef HAVE_RESOURCE
		resource_set_acquire (priv->resource);
#else
                /* There's no callback so go directly */
		thumber_set_state (thumber, THUMBER_STATE_WORKING);
#endif
	}
}

static void
release_resources (Thumber *thumber)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (thumber);

	if (priv->state == THUMBER_STATE_PENDING_STOP ||
            priv->state == THUMBER_STATE_STOPPED) {
#ifdef HAVE_RESOURCE
		resource_set_release (priv->resource);
#else
                /* There's no callback so go directly */
		thumber_set_state (thumber, THUMBER_STATE_STOPPED);
#endif
	} else {
#ifdef HAVE_RESOURCE
		resource_set_release (priv->resource);
#else
                /* There's no callback so go directly */
		thumber_set_state (thumber, THUMBER_STATE_PAUSED);
#endif               
        }
}


gboolean
thumber_dbus_register(Thumber *thumber,
                      const gchar *bus_name,
                      const gchar *bus_path,
                      GError **error)
{
	GError          *lerror     = NULL;
	DBusGConnection *connection = NULL;
	DBusGProxy      *proxy      = NULL;
	guint            result;
	ThumberPrivate  *priv;
	priv = THUMBER_GET_PRIVATE (thumber);

	connection = dbus_g_bus_get (DBUS_BUS_SESSION, &lerror);

	if (lerror) {
		g_printerr ("Problem with dbus, %s\n", lerror->message);
		g_error_free (lerror);
		return FALSE;
	}

	proxy = dbus_g_proxy_new_for_name (connection,
	                                   DBUS_SERVICE_DBUS,
	                                   DBUS_PATH_DBUS,
	                                   DBUS_INTERFACE_DBUS);

	org_freedesktop_DBus_request_name (proxy, bus_name,
	                                   DBUS_NAME_FLAG_DO_NOT_QUEUE,
	                                   &result, &lerror);

	g_object_unref (proxy);

	if (lerror) {
		g_printerr ("Problem requesting name, %s\n", lerror->message);
		g_error_free (lerror);
		return FALSE;
	}

	dbus_g_object_type_install_info (G_OBJECT_TYPE (thumber),
	                                 &dbus_glib_gst_video_thumbnailer_object_info);

	dbus_g_connection_register_g_object (connection,
	                                     bus_path,
	                                     G_OBJECT (thumber));

#ifdef HAVE_RESOURCE
	/* resource control */
	priv->resource = resource_set_create("background",
	                                     RESOURCE_VIDEO_PLAYBACK,
	                                     0,
	                                     RESOURCE_ALWAYS_REPLY,
	                                     resource_grant_handler,
	                                     thumber);

	resource_set_configure_advice_callback (priv->resource,
	                                        resource_advice_handler,
	                                        thumber);
#endif



	return TRUE;
}

static void
add_file (Thumber *thumber,
          FileInfo *info)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (thumber);

	g_queue_push_tail (priv->file_queue, info);

}

static void
add_task (Thumber *thumber,
          TaskInfo *info)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (thumber);

	g_queue_push_tail (priv->task_queue, info);
}

static gboolean
quit_timeout_cb (gpointer user_data)
{
	gst_thumb_main_quit ();
	return FALSE;
}

static void
thumber_set_state (Thumber *thumber,
                   ThumberState state)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (thumber);

	switch (state) {
		case THUMBER_STATE_WORKING:
			if (priv->idle_id == 0) {
				priv->idle_id = g_idle_add (thumber_process_func,
				                            thumber);
			}
			if (priv->quit_timeout_id != 0) {
				g_source_remove (priv->quit_timeout_id);
				priv->quit_timeout_id = 0;
			}
			break;
		case THUMBER_STATE_PAUSED:
			if (priv->idle_id != 0) {
				g_source_remove (priv->idle_id);
				priv->idle_id = 0;
			}
			if (priv->quit_timeout_id != 0) {
				g_source_remove (priv->quit_timeout_id);
				priv->quit_timeout_id = 0;
			}
			break;
		case THUMBER_STATE_PENDING_STOP:
			if (priv->idle_id != 0) {
				g_source_remove (priv->idle_id);
				priv->idle_id = 0;
			}
			if (priv->quit_timeout_id == 0) {
				if (priv->quit_timeout >= 0) {
					priv->quit_timeout_id = g_timeout_add_seconds (priv->quit_timeout,
					                                               quit_timeout_cb,
					                                               thumber);
				}
			}
			break;
		case THUMBER_STATE_STOPPED:
			if (priv->idle_id != 0) {
				g_source_remove (priv->idle_id);
				priv->idle_id = 0;
			}
			if (priv->quit_timeout_id == 0) {
				priv->quit_timeout_id = g_timeout_add_seconds (priv->quit_timeout,
				                                               quit_timeout_cb,
				                                               thumber);
			}
			break;
		default:
			/* Nothing */
			break;
	}

	priv->state = state;
}

static void
thumber_start (Thumber *object)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (object);

        if (priv->state == THUMBER_STATE_STOPPED ||
            priv->state == THUMBER_STATE_PENDING_STOP) {
            thumber_set_state (object, THUMBER_STATE_PAUSED);
            request_resources (object);
        }
}

static void
thumber_pause (Thumber *object)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (object);

        if (priv->state == THUMBER_STATE_WORKING) {
                thumber_set_state (object, THUMBER_STATE_PAUSED);
                release_resources (object);
        }
}

static void
thumber_stop (Thumber *object)
{
	ThumberPrivate *priv;
	priv = THUMBER_GET_PRIVATE (object);

        if (priv->state == THUMBER_STATE_WORKING) {
                thumber_set_state (object, THUMBER_STATE_PENDING_STOP);
                release_resources (object);
        }
}

void
thumber_dbus_method_queue (Thumber *object,
                           gchar *uri,
                           gchar *mime_hint,
                           gchar *flavor,
                           gboolean priority,
                           DBusGMethodInvocation *context)
{
	TaskInfo    *task   = NULL;
	const ThumberFlavor *flav = thumber_get_flavor (object, flavor);

	if (!flav) {
		GError *error = NULL;
		g_set_error (&error,
		             error_quark (),
		             DBUS_ERROR,
		             "Could not find flavor %s", flavor);
		dbus_g_method_return_error (context, error);
		g_error_free (error);
		return;
	}

	task = task_info_new (mime_hint);
	task_info_add_uri (task, uri, flav);

	add_task (object, task);

        thumber_start (object);

	dbus_g_method_return (context, task->id);
}



void
thumber_dbus_method_dequeue (Thumber *object,
                             guint handle,
                             DBusGMethodInvocation *context)
{
	/* Not supported ATM */
	dbus_g_method_return (context);
}


void
thumber_populate_file_queue (Thumber *thumber, TaskInfo *task) {

	GSList *list = NULL;

	list = task->files;

	for (list = task->files; list; list = g_slist_next (list)) {
		FileInfo *info;
		info = list->data;

		add_file (thumber, info);
	}
}

static gboolean
thumber_process_func (gpointer data)
{
	Thumber *thumber;
	FileInfo *file;
	TaskInfo *task;
	ThumberPrivate *priv;
	GError *error = NULL;
	thumber = THUMBER (data);
	priv = THUMBER_GET_PRIVATE (thumber);

	if ((file = g_queue_peek_head (priv->file_queue)) != NULL) {
		
		if (!thumber_pipe_run (priv->pipe,
				       file->uri,
				       file->flavor,
				       &error)) {
			
			if (error) {
                                /* Check for dsp resource preemption and treat that as "try later".
                                 * We just stop processing and hope that we will get the matching
                                 * policy commands with a delay.
                                 * FIXME? This is evil in many ways but what can you do... */
                                
                                if (error->domain == g_quark_from_string("gst-resource-error-quark")) {
                                        g_warning ("Thumbnailing was preempted for %s. Will try again later: %s",
                                                   file->uri,
                                                   error->message);
                                        thumber_pause (thumber);
                                        /* We don't want to remove the file from queue */
                                        return FALSE;
                                }

				g_warning ("Failed to run thumbnailing pipeline for %s, error: %s",
					   file->uri, error->message);
				g_signal_emit (thumber,
					       signals[ERROR_SIGNAL],
					       0,
					       file->id,
					       file->uri,
					       1,
					       error->message);
				g_error_free (error);
				error = NULL;
			} else {
                                g_signal_emit (thumber,
					       signals[ERROR_SIGNAL],
					       0,
					       file->id,
					       file->uri,
					       0,
					       "Undefined error");
			}
		} else {
		        g_signal_emit (thumber,
				       signals[READY_SIGNAL],
				       0,
				       file->id,
				       file->uri);
		}
		
		file = g_queue_pop_head (priv->file_queue);
		file_info_free (file);
		
                if ((file = g_queue_peek_head (priv->file_queue)) == NULL) {
                        if (priv->current_task) {
                                g_signal_emit (thumber,
                                               signals[FINISHED_SIGNAL],
                                               0,
                                               priv->current_task->id);
                                task_info_free (priv->current_task, FALSE);
                                priv->current_task = NULL;
                        }
                }
	} else {
		if (priv->current_task) {
			g_signal_emit (thumber,
			               signals[FINISHED_SIGNAL],
			               0,
			               priv->current_task->id);
			task_info_free (priv->current_task, FALSE);
			priv->current_task = NULL;
		}
		
		if ((task = g_queue_pop_head (priv->task_queue)) != NULL) {

			thumber_populate_file_queue (thumber, task);
			
			priv->current_task = task;
			g_signal_emit (thumber,
			               signals[STARTED_SIGNAL],
			               0,
			               priv->current_task->id);
		} else {
			
			thumber_stop (thumber);
			return FALSE;
		}
	}

	return TRUE;
}

GList *
thumber_get_flavors (Thumber     *thumber)
{
	GList           *flavors = NULL;
	GList           *iter;
	ThumberPrivate  *priv;

	priv = THUMBER_GET_PRIVATE (thumber);

	for (iter = g_list_last (priv->flavors); iter != NULL; iter = iter->prev)
		flavors = g_list_prepend (flavors, g_object_ref (iter->data));

	return flavors;
}

const ThumberFlavor *
thumber_get_flavor (Thumber     *thumber,
                    const gchar *flavor)
{
	GList           *iter;
	ThumberPrivate  *priv;

	priv = THUMBER_GET_PRIVATE (thumber);

	for (iter = g_list_last (priv->flavors); iter != NULL; iter = iter->prev) {
		if (strcmp (thumber_flavor_get_name (iter->data), flavor) == 0) {
			return iter->data;
		}
	}

	return NULL;
}

static void
load_flavors (Thumber *thumber)
{
	ThumberFlavor          *flavor;
	GKeyFile               *key_file;
	GError                 *error = NULL;
	gchar                 **sections;
	gchar                  *filename = NULL;
	gint                    width;
	gint                    height;
	gint                    min_width;
	gint                    min_height;
	gchar                  *name;
	gchar                  *frame_selection;
        gchar                  *rotation;
	guint                   n;
	const gchar * const    *config_dirs;
	ThumberPrivate         *priv;

	priv = THUMBER_GET_PRIVATE (thumber);

	config_dirs = g_get_system_config_dirs ();

	for (n = 0; config_dirs[n] != NULL; n++) {
		gchar *filen;

		filen = g_build_filename (config_dirs[n], "thumbnails",
		                          "flavors.conf", NULL);

		if (g_file_test (filen, G_FILE_TEST_IS_REGULAR)) {
			filename = filen;
			break;
		} else {
			g_free (filen);
		}
	}

	if (!filename) {
		flavor = thumber_flavor_new_normal ();
		priv->flavors = g_list_prepend (priv->flavors, flavor);

		flavor = thumber_flavor_new_large ();
		priv->flavors = g_list_prepend (priv->flavors, flavor);

		return;
	}

	/* allocate the key file */
	key_file = g_key_file_new ();

	/* try to load the key file from the overrides file */
	if (!g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, &error)) {
		g_warning ("Failed to load the file \"%s\": %s", filename, error->message);
		g_clear_error (&error);
		g_key_file_free (key_file);
		g_free (filename);

		flavor = thumber_flavor_new_normal ();
		priv->flavors = g_list_prepend (priv->flavors, flavor);

		flavor = thumber_flavor_new_large ();
		priv->flavors = g_list_prepend (priv->flavors, flavor);

		return;
	}

	/* determine sections in the key file */
	sections = g_key_file_get_groups (key_file, NULL);

	/* iterate over all sections */
	for (n = 0; sections != NULL && sections[n] != NULL; ++n) {

		name = g_key_file_get_string (key_file, sections[n], "Name", &error);

		if (error != NULL) {
			g_warning ("Malformed section \"%s\" in file \"%s\": %s",
			           sections[n], filename, error->message);
			g_clear_error (&error);
			continue;
		}

		width = g_key_file_get_integer (key_file, sections[n], "Width", &error);

		if (error != NULL) {
			g_warning ("Malformed section \"%s\" in file \"%s\": %s",
			           sections[n], filename, error->message);
			g_clear_error (&error);
			continue;
		}

		height = g_key_file_get_integer (key_file, sections[n], "Height", &error);

		if (error != NULL) {
			g_warning ("Malformed section \"%s\" in file \"%s\": %s",
			           sections[n], filename, error->message);
			g_clear_error (&error);
			continue;
		}

		min_width = g_key_file_get_integer (key_file, sections[n], "MinimumWidth", &error);

		if (error != NULL) {
			if (error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND) {
				g_warning ("Malformed section \"%s\" in file \"%s\": %s",
				           sections[n], filename, error->message);
				g_clear_error (&error);
				continue;
			} else {
				g_clear_error (&error);
			}
		}

		if (min_width > width) {
			g_warning ("Error in section \"%s\" in file \"%s\": minimum width %d greater than maximum %d. Ignoring.",
			           sections[n], filename, min_width, width);
			min_width = 0;
		}

		min_height = g_key_file_get_integer (key_file, sections[n], "MinimumHeight", &error);

		if (error != NULL) {
			if (error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND) {
				g_warning ("Malformed section \"%s\" in file \"%s\": %s",
				           sections[n], filename, error->message);
				g_clear_error (&error);
				continue;
			} else {
				g_clear_error (&error);
			}
		}

		if (min_height > height) {
			g_warning ("Error in section \"%s\" in file \"%s\": minimum height %d greater than maximum %d. Ignoring.",
			           sections[n], filename, min_height, height);
			min_width = 0;
		}

		frame_selection = g_key_file_get_string (key_file, sections[n], "FrameSelectionPolicy", &error);

		if (error != NULL) {
			if (error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND) {
				g_warning ("Malformed section \"%s\" in file \"%s\": %s",
				           sections[n], filename, error->message);
				g_clear_error (&error);
				continue;
			} else {
				g_clear_error (&error);
			}

			frame_selection = g_strdup ("Normal");
		}

		rotation = g_key_file_get_string (key_file, sections[n], "RotationPolicy", &error);

		if (error != NULL) {
			if (error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND) {
				g_warning ("Malformed section \"%s\" in file \"%s\": %s",
				           sections[n], filename, error->message);
				g_clear_error (&error);
				continue;
			} else {
				g_clear_error (&error);
			}

			rotation = g_strdup ("Rotate");
		}

		flavor = thumber_flavor_new (name, width, height, min_width, min_height, frame_selection, rotation);
		priv->flavors = g_list_prepend (priv->flavors, flavor);

		g_free (name);
		g_free (frame_selection);
	}

	g_key_file_free (key_file);
	g_strfreev (sections);
	g_free (filename);
}
